@@ -3,8 +3,6 @@
 BLOCK_SIZE = 1048576 * 4
 
 import os, sys, hashlib, lzma, aes, time, zlib, bz2, email, types
-import email.mime.multipart
-import email.mime.application
 import cPickle as pickle
 import adler64
 from itertools import ifilter
@@ -71,7 +69,7 @@ class stor_stream:
   def open(self, backend):
     self._backend = backend
 
-  def _checksum(self, data):
+  def checksum(self, data):
     ctx = hashlib.sha256()
     if self._pass != None: ctx.update(self._pass)
     ctx.update(data)
@@ -83,7 +81,7 @@ class stor_stream:
   def archive(self):
     return self._archive
 
-  def append_block_info(self, number, checksum1, checksum2):
+  def append_block_info(self, number, checksum1, checksum2, backend):
     checksum1 = int(checksum1, 16)
     if number == 0 and checksum2 not in self._known_metadata: self._known_metadata.append(checksum2)
     elif checksum1 == 0: self._known_parts.setdefault(checksum2, []).append(number)
@@ -93,23 +91,13 @@ class stor_stream:
     self._backend.close()
 
   def _decode_message(self, buffer, checksum, bufflen, meta = False):
-    if self._backend.should_mime():
-      messagedata = email.message_from_string(buffer)
-      message = None
-      for part in messagedata.walk():
-          if part.get_content_type() != 'application/octet-stream': continue
-          message = part.get_payload(decode = True)
-          break
-      if message == None:
-        raise SystemError("Data has wrong structure.")
-    else: message = buffer
-    data = message
+    data = buffer
     try:
-      if self._aes: data = self._aes.decrypt(message)
+      if self._aes: data = self._aes.decrypt(data)
       if meta: data = bz2.decompress(data)
       else: data = self._decompress(data)
       if bufflen != 0 and len(data) > bufflen: data = data[:bufflen]
-      if self._checksum(data) != checksum:
+      if self.checksum(data) != checksum:
 	raise SystemError()
     except: raise SystemError("Wrong password or invalid data")
     return data
@@ -193,7 +181,7 @@ class stor_output(stor_stream):
 
   def commit_buffer(self, buffer, checksum1int, meta = False):
     checksum1 = "%016X" % checksum1int
-    checksum2 = self._checksum(buffer)
+    checksum2 = self.checksum(buffer)
     buflen = len(buffer)
     if checksum1int == 0 and self._known_parts.has_key(checksum2):
       self._metadata['blocks'].append((self._known_parts[checksum2], checksum1, checksum2, buflen))
@@ -205,13 +193,13 @@ class stor_output(stor_stream):
 	buffer = buffer + "\0" * (16 - bufrest)
       if self._aes:
 	buffer = self._aes.encrypt(buffer)
-      if self._backend.should_mime():
-	message = email.mime.multipart.MIMEMultipart()
-	message.attach(email.mime.application.MIMEApplication(buffer))
-	message = message.as_string()
-      else: message = buffer
-      self._size += len(message)
-      self._backend.write_buffer(message, self._bufnum, checksum1, checksum2)
+      #if self._backend.should_mime():
+      #  message = email.mime.multipart.MIMEMultipart()
+      #  message.attach(email.mime.application.MIMEApplication(buffer))
+      #  message = message.as_string()
+      #else: message = buffer
+      self._size += len(buffer)
+      self._backend.write_buffer(buffer, self._bufnum, checksum1, checksum2)
       self._metadata['blocks'].append((self._bufnum, checksum1, checksum2, buflen))
     if self._bufnum != 0: self._bufnum += 1
     else: self._archive = checksum2
@@ -229,10 +217,9 @@ class stor_input(stor_stream):
     stor_stream.open(self, backend)
     if self._archive_checksum not in self._known_metadata:
       raise SystemError("Archive %s not found." % repr(self._archive_checksum))
-    metadata_buffer = self._backend.read_buffer(0, "%016X" % 0, self._archive_checksum)
+    self._metadata = self._backend.read_buffer(0, "%016X" % 0, self._archive_checksum,
+	lambda buf: pickle.loads(self._decode_message(buf, self._archive_checksum, 0, meta = True)))
     self._decompress = lambda buffer: bz2.decompress(buffer)
-    metadata_buffer = self._decode_message(metadata_buffer, self._archive_checksum, 0, meta = True)
-    self._metadata = pickle.loads(metadata_buffer)
     if self._metadata['compression'][0] == 'lzma': self._decompress = lambda buffer: lzma.decompress(buffer)
     elif self._metadata['compression'][0] == 'gzip': self._decompress = lambda buffer: zlib.decompress(buffer)
     elif self._metadata['compression'][0] == 'bzip2': self._decompress = lambda buffer: bz2.decompress(buffer)
@@ -243,7 +230,7 @@ class stor_input(stor_stream):
   def read(self, size):
     while len(self._buffer) < size and len(self._blocks) > 0:
       num, ck1, ck2, blen = self._blocks.pop(0)
-      data = self._decode_message(self._backend.read_buffer(num, ck1, ck2), ck2, blen)
+      data = self._backend.read_buffer(num, ck1, ck2, lambda buf: self._decode_message(buf, ck2, blen))
       self._buffer = self._buffer + data
       self._bufferlen += len(data)
     data = self._buffer[:size]
@@ -253,17 +240,19 @@ class stor_input(stor_stream):
   def metainfo(self):
     return self._metadata['metainfo']
 
-class stor_search(stor_stream):
+class stor_util(stor_stream):
   def __init__(self, password):
     stor_stream.__init__(self, password)
 
   def search(self, keywords):
-    result = []; parents = {}
+    result = []; parents = {}; errors = []
     checksum1 = "%016X" % 0
     for metadata in self._known_metadata:
-      data = self._backend.read_buffer(0, checksum1, metadata)
-      data = self._decode_message(data, metadata, 0, meta = True)
-      data = pickle.loads(data)
+      try: data = self._backend.read_buffer(0, checksum1, metadata,
+            lambda buf: pickle.loads(self._decode_message(buf, metadata, 0, meta = True)))
+      except Exception, err:
+	errors.append((metadata, err))
+	continue
       if len(keywords) > 0:
         weight = 0
         for kwd in keywords:
@@ -295,6 +284,20 @@ class stor_search(stor_stream):
       parents.setdefault(data['parent'], []).append(metadata)
     if len(keywords) > 0: result.sort(key = lambda x: x['weight'])
     else: result.sort(key = lambda x: x['description'])
-    return (result, parents)
+    return (result, parents, errors)
 
-__all__ = ('stor_output', 'stor_input')
+  def keywords(self):
+    checksum1 = "%016X" % 0
+    keywords = {}
+    errors = []
+    for metadata in self._known_metadata:
+      try: data = self._backend.read_buffer(0, checksum1, metadata,
+            lambda buf: pickle.loads(self._decode_message(buf, metadata, 0, meta = True)))
+      except Exception, err:
+	errors.append((metadata, err))
+	continue
+      for kwd in data['keywords']:
+	keywords[kwd] = keywords.get(kwd, 0) + 1
+    return (keywords, errors)
+
+__all__ = ('stor_output', 'stor_input', 'stor_util')
